# $Id: record.py,v 1.52 2011/05/02 01:21:17 irees Exp $
import time
import UserDict
import collections
import operator
import weakref
import copy
import re

import emen2.db.btrees
import emen2.db.datatypes
import emen2.db.exceptions
import emen2.db.dataobject
import emen2.util.listops as listops

import emen2.db.config
g = emen2.db.config.g()


class Record(emen2.db.dataobject.PermissionsDBObject):
	"""This class encapsulates a single database record. In a sense this is an instance
	of a particular RecordDef, however, note that it is not required to have a value for
	every field described in the RecordDef, though this will usually be the case.

	To modify the params in a record use the normal obj[key]= or update() approaches.
	Changes are not stored in the database until commit() is called. To examine params,
	use obj[key]. There are a few special keys, handled differently:
	creator, creationtime, permissions, comments, history, parents, children, groups

	Record instances must ONLY be created by the Database class through retrieval or
	creation operations. self._ctx will store information about security and
	storage for the record.

	Mechanisms for changing existing params are a bit complicated. In a sense, as in a
	physical lab notebook, an original value can never be changed, only superceded.
	All records have a 'magic' field called 'comments', which is an extensible array
	of text blocks with immutable entries. 'comments' entries can contain new field
	definitions, which will supercede the original definition as well as any previous
	comments. Changing a field will result in a new comment being automatically generated
	describing and logging the value change.

	From a database standpoint, this is rather odd behavior. Such tasks would generally be
	handled with an audit log of some sort. However, in this case, as an electronic
	representation of a Scientific lab notebook, it is absolutely necessary
	that all historical values are permanently preserved for any field, and there is no
	particular reason to store this information in a separate file. Generally speaking,
	such changes should be infrequent.

	Naturally, as with anything in Python, anyone with code-level access to the database
	can override this behavior by changing 'params' directly rather than using
	the supplied access methods. There may be appropriate uses for this when constructing
	a new Record before committing changes back to the database.
	"""


	# Protected params are history, "real" comments, rectype, and "_params"
	param_all = emen2.db.dataobject.PermissionsDBObject.param_all | set(['comments', 'history', 'rectype'])
	param_required = set(['rectype'])
	recid = property(lambda s:s.name)

	def init(self, d):			
		super(Record, self).init(d)
		
		# rectype is required
		self.__dict__['rectype'] = d.pop('rectype')

		# comments, history, and other param values
		self.__dict__['comments'] = []
		self.__dict__['history'] = []		
		self.__dict__['params'] = {}

		# Access to RecordDef is checked during validate


	#################################
	# repr methods
	#################################

	def __repr__(self):
		return "<%s name: %s at %x>" % (self.__class__.name, self.name, id(self))


	def json_equivalent(self):
		"""Returns a dictionary of current values"""
		ret = {}
		ret.update(self.params)
		for i in self.param_all:
			ret[i] = getattr(self, i, None)
		return ret


	def __repr__(self):
		return "<%s %s, %s at %x>" % (self.__class__.__name__, self.name, self.rectype, id(self))


	#################################
	# Setters
	#################################

	def _set_comments(self, key, value, vtm=None, t=None):
		return self.addcomment(value, t=t)


	# in Record, params not in self.param_all are put in self.params{}.
	def _setoob(self, key, value, vtm=None, t=None):		
		if self.params.get(key) == value:
			# print ":: No change: ", key, value
			return set()
		self.vw(key) # Check permissions
		self._addhistory(key, t=t)
		# print ":: Setting param key/value", key, value
		self.params[key] = value
		return set([key])


	#################################
	# mapping methods;
	#		UserDict.DictMixin provides the remainder
	#################################

	def __getitem__(self, key, default=None):
		"""Default behavior is similar to .get: return None as default"""
		if key in self.param_all:
			return getattr(self, key, default)
		else:		
			return self.params.get(key, default)


	def keys(self):
		"""All retrievable keys for this record"""
		return self.params.keys() + list(self.param_all)


	def getparamkeys(self):
		"""Returns parameter keys without special values like owner, creator, etc."""
		return self.params.keys()



	##########################
	# Comments and history
	##########################

	def _addhistory(self, param, t=None):
		# Changes aren't logged on uncommitted records
		if self.name < 0:
			return

		if not param:
			raise Exception, "Unable to add item to history log"

		vtm, t = self._vtmtime(t=t)			
		self.history.append((unicode(self._ctx.username), unicode(t), unicode(param), self.params.get(param)))


	def addcomment(self, value, vtm=False, t=None):
		self.vw('comments', self.commentable(), 'Insufficient permissions to add comment')

		vtm, t = self._vtmtime(vtm, t)
		cp = set()
		if value == None:
			return set()
		
		newcomments = []	
		if hasattr(value, "__iter__"):
			check = [(unicode(i[0]), unicode(i[1]), unicode(i[2])) for i in value]
			existing = [(unicode(i[0]), unicode(i[1]), unicode(i[2])) for i in self.comments]
			for c in check:
				if c not in existing:
					newcomments.append(c[2])
		else:
			newcomments.append(unicode(value))

		# newcomments2 = []
		# updvalues = {}		
		for value in newcomments:
			d = {}
			if not value.startswith("LOG"): # legacy fix..
				d = emen2.db.recorddef.parseparmvalues(value)[1]

			if d.has_key("comments"):
				# Always abort
				self.error("Cannot set comments inside a comment", warning=False)
						
			# Now update the values of any embedded params
			for i,j in d.items():
				cp |= self.__setitem__(i, j, vtm=vtm, t=t)

			# Store the comment string itself
			self.comments.append((unicode(self._ctx.username), unicode(t), unicode(value)))
			cp.add('comments')
		
		return cp


	def revision(self, revision=None):
		"""This will return information about the record's revision history"""

		history = copy.copy(self.history)
		comments = copy.copy(self.comments)
		comments.append((self.get('creator'), self.get('creationtime'), 'Created'))
		paramcopy = {}

		bydate = collections.defaultdict(list)
		
		for i in filter(lambda x:x[1]>=revision, history):
			bydate[i[1]].append([i[0], i[2], i[3]])

		for i in filter(lambda x:x[1]>=revision, comments):
			bydate[i[1]].append([i[0], None, i[2]])
			
		revs = sorted(bydate.keys(), reverse=True)

		for rev in revs:				
			for item in bydate.get(rev, []):
				# user, param, oldval
				if item[1] == None:
					continue
				if item[1] in paramcopy.keys():
					newval = paramcopy.get(item[1])
				else:
					newval = self.get(item[1])
	
				paramcopy[item[1]] = copy.copy(item[2])				
				# item[2] = newval

		return bydate, paramcopy
		


	#################################
	# validation methods
	#################################

	
	def validate_name(self, name):
		"""Validate the name of this object"""		
		if name in ['None', None]:
			return
		try:
			name = int(name)			
		except ValueError:
			self.error("Name must be an integer")
		return name



	def validate(self, vtm=None, t=None):
		# Cut out any None's
		pitems = self.params.items()
		for k,v in pitems:
			if not v and v != 0 and v != False:
				del self.params[k]
		
		# Check the rectype and any required parameters
		vtm, t = self._vtmtime(vtm=vtm, t=t)
		cachekey = vtm.get_cache_key('recorddef', self.rectype)
		hit, rd = vtm.check_cache(cachekey)
		if not hit:
			rd = self._ctx.db.getrecorddef(self.rectype, filt=False)
			vtm.store(cachekey, rd)
	
		for param in rd.paramsR:
			if self.get(param) == None:
				self.error("%s is a required parameter"%(param))
	
		self.__dict__['rectype'] = unicode(rd.name)
		




class RecordDB(emen2.db.btrees.RelateDB):
	sequence = True
	cfunc = False
	keytype = 'd'
	dataclass = Record

	def openindex(self, param, txn=None):
		# Parents / children
		ind = super(RecordDB, self).openindex(param, txn=txn)
		if ind:
			return ind

		# print "Attempting to open index %s"%param
		# RecordDB maintains an index-of-indexes for faster querying
		if param == 'indexkeys':
			ind = emen2.db.btrees.IndexKeysDB(filename=self._indname(param), keytype='s', datatype='p', dbenv=self.dbenv)
			return ind
			
		# Check the paramdef to see if it's indexed
		# ian: todo: use the context
		# pd = ctx.getparamdef(param, filt=False)
		pd = self.dbenv.paramdef.get(param, filt=False, txn=txn)
		if not pd or pd.vartype not in self.dbenv.indexablevartypes or not pd.indexed:
			return None

		# disable this check, and always create index.			
		vtm = emen2.db.datatypes.VartypeManager()
		tp = vtm.getvartype(pd.vartype).keytype

		ind = emen2.db.btrees.IndexDB(filename=self._indname(param), keytype=tp, datatype='d', dbenv=self.dbenv)
		return ind


	# Update the database sequence.. Probably move this to the parent class.
	def update_sequence(self, items, txn=None):
		# Which recs are new?
		newrecs = [i for i in items if i.name < 0]
		namemap = {}

		# Reassign new record IDs and update record counter
		if newrecs:
			basename = self._set_sequence(delta=len(newrecs), txn=txn)

		# We have to manually update the rec.__dict__['name'] because this is normally considered a reserved attribute.
		for offset, newrec in enumerate(newrecs):
			oldname = newrec.name
			newrec.__dict__['name'] = offset + basename
			namemap[oldname] = newrec.name		

		# Update all the record's links
		for item in items:
			item.parents = set([namemap.get(i,i) for i in item.parents])
			item.children = set([namemap.get(i,i) for i in item.children])

		return namemap
	
	
	
	def delete(self, names, ctx=None, txn=None):
		recs = self.cgets(names, ctx=ctx, txn=txn)
		crecs = []
		for rec in recs:
			rec.setpermissions([[],[],[],[]])
			if rec.parents and rec.children:
				rec["comments"] = "Record hidden by unlinking from parents %s and children %s"%(", ".join([unicode(x) for x in rec.parents]), ", ".join([unicode(x) for x in rec.children]))
			elif rec.parents:
				rec["comments"] = "Record hidden by unlinking from parents %s"%", ".join([unicode(x) for x in rec.parents])
			elif rec.children:
				rec["comments"] = "Record hidden by unlinking from children %s"%", ".join([unicode(x) for x in rec.children])
			else:
				rec["comments"] = "Record hidden"
			
			rec['deleted'] = True
			rec['children'] = set()
			rec['parents'] = set()
			crecs.append(rec)
		
		return self.cputs(crecs, ctx=ctx, txn=txn)	



	def groupbyrectype(self, names, ctx=None, txn=None):
		"""Group Records by Rectype. Filters for permissions.	
		@param names Record(s) or Record name(s)
		@return {rectype:set(record names)}
		"""
		if not names:
			return {}

		# Allow either Record(s) or Record name(s) as input
		ret = collections.defaultdict(set)
		recnames, recs, other = listops.typepartition(names, int, emen2.db.dataobject.BaseDBObject)

		if len(recnames) < 1000:
			# Just get the rest of the records directly
			recs.extend(self.cgets(recnames, ctx=ctx, txn=txn))
		else:
			# Use the index for large numbers of records
			ind = self.getindex("rectype", txn=txn)
			# Filter permissions
			names = self.filter(recnames, ctx=ctx, txn=txn)
			while names:
				# get a random record's rectype
				rid = names.pop() 
				rec = self.get(rid, txn=txn) 
				# get the set of all records with this recorddef
				ret[rec.rectype] = ind.get(rec.rectype, txn=txn) & names 
				# remove the results from our list since we have now classified them
				names -= ret[rec.rectype] 
				# add back the initial record to the set
				ret[rec.rectype].add(rid)

		for i in recs:
			ret[i.rectype].add(i.name)
		
		return ret
		
	
	# This builds UP instead of prunes DOWN; filter does the opposite..
	def names(self, names=None, ctx=None, txn=None, **kwargs):
		if names is not None:
			return self.filter(names, rectype=kwargs.get('rectype'), ctx=ctx, txn=txn)

		if ctx.checkreadadmin():
			return set(xrange(self.get_max(txn=txn)))

		ind = self.getindex("permissions", txn=txn)
		indg = self.getindex("groups", txn=txn)
		ret = ind.get(ctx.username, set(), txn=txn)
		for group in sorted(ctx.groups, reverse=True):
			ret |= indg.get(group, set(), txn=txn)

		return ret
		
		
	def filter(self, names, rectype=None, ctx=None, txn=None):
		"""Filter for permissions.
		@param names Record name(s).		
		@return Readable Record names.
		"""
		
		names = self.expand(names, ctx=ctx, txn=txn)

		if rectype:
			ind = self.getindex('rectype', txn=txn)
			# ian: use the context.
			rd = set()
			for i in ctx.db.getrecorddef(listops.check_iterable(rectype)):
				rd |= ind.get(i.name, txn=txn)
			names &= rd

		if ctx.checkreadadmin():
			return names

		# ian: indexes are now faster, generally...
		if len(names) <= 1000:
			crecs = self.cgets(names, ctx=ctx, txn=txn)
			return set([i.name for i in crecs])

		# Make a copy
		find = copy.copy(names)

		# Use the permissions/groups index
		ind = self.getindex('permissions', txn=txn)
		indg = self.getindex('groups', txn=txn)

		find -= ind.get(ctx.username, set(), txn=txn)
		for group in sorted(ctx.groups):
			if find:
				find -= indg.get(group, set(), txn=txn)

			
		return names - find


		
	

__version__ = "$Revision: 1.52 $".split(":")[1][:-1].strip()
